@@ -103,6 +103,9 @@ class Transaction(object):
         else:
             self.data = base64url_encode(data.encode('ascii'))
 
+        if self.data is None:
+            self.data = ''
+
         self.file_handler = kwargs.get('file_handler', None)
         if self.file_handler:
             self.uses_uploader = True
@@ -184,7 +187,7 @@ class Transaction(object):
     def get_signature_data(self):
         self.reward = self.get_reward(self.data_size, target_address=self.target if len(self.target) > 0 else None)
 
-        if self.data_size > 0 and self.data_root == "":
+        if self.data_size > 0 and self.data_root == "" and not self.uses_uploader:
             if type(self.data) == str:
                 root_hash = compute_root_hash(io.StringIO(self.data))
 
@@ -248,8 +251,12 @@ class Transaction(object):
         return self.last_tx
 
     def to_dict(self):
+
+        if self.data is None:
+            self.data = ''
+
         data = {
-            'data': self.data.decode(),
+            'data': self.data.decode() if type(self.data) == bytes else self.data,
             'id': self.id.decode() if type(self.id) == bytes else self.id,
             'last_tx': self.last_tx,
             'owner': self.owner,
